import unittest

import numpy as np
import pandas as pd
import xarray as xr

from tapr.main.conversion import ntable, tabulate
from tapr.main.ntable import NTable
from tests.testing_utils import assert_ntable_equivalent


class TestNTable(unittest.TestCase):
    def setUp(self):
        self._dictionary = {
            "x1": {"y1": 0, "y2": "1", "y3": 2},
            "x2": {"y1": "three", "y2": 4, "y3": "5"},
        }

        self._dataframe = pd.DataFrame(self._dictionary)
        self._dataarray = xr.DataArray(
            self._dataframe.to_numpy().T,
            coords={"dim0": ["x1", "x2"], "dim1": ["y1", "y2", "y3"]},
            dims=["dim0", "dim1"],
        )
        reflist = [0, "1", 2, "three", 4, "5"]
        refmap = xr.DataArray(
            np.arange(6).reshape((2, 3)),
            coords={"dim0": ["x1", "x2"], "dim1": ["y1", "y2", "y3"]},
            dims=["dim0", "dim1"],
        )
        self._expected_ntbl = NTable(reflist, refmap)

    def test_ntable_dictionary(self):
        ntbl = ntable(self._dictionary)
        assert_ntable_equivalent(self._expected_ntbl, ntbl)

    def test_ntable_dataframe(self):
        ntbl = ntable(self._dataframe, dims=("dim1", "dim0")).struct.T
        assert_ntable_equivalent(self._expected_ntbl, ntbl)

    def test_ntable_dataarray(self):
        ntbl = ntable(self._dataarray)
        assert_ntable_equivalent(self._expected_ntbl, ntbl)

    def test_ntable_ntable(self):
        ntbl = ntable(self._expected_ntbl)
        assert_ntable_equivalent(self._expected_ntbl, ntbl)

    def test_ntable_collection(self):
        ntbl = ntable((self._expected_ntbl, 3))
        reflist = [(0, 3), ("1", 3), (2, 3), ("three", 3), (4, 3), ("5", 3)]
        refmap = xr.DataArray(
            np.arange(6).reshape((2, 3)),
            coords={"dim0": ["x1", "x2"], "dim1": ["y1", "y2", "y3"]},
            dims=["dim0", "dim1"],
        )
        expected_ntbl = NTable(reflist, refmap)
        assert_ntable_equivalent(expected_ntbl, ntbl)

    def test_ntable_unable(self):
        result = False
        test = 30
        try:
            blah = ntable(test)
        except TypeError:
            result = True
        self.assertTrue(result)


class TestTabulate(unittest.TestCase):
    def setUp(self):
        reflist = [0, "1", 2, "three", 4, "5"]
        self._refmap = xr.DataArray(
            np.arange(6).reshape((2, 3)),
            coords={"dim0": ["x1", "x2"], "dim1": ["y1", "y2", "y3"]},
            dims=["dim0", "dim1"],
        )
        self._ntbl = NTable(reflist, self._refmap)
        self._tuple = (self._ntbl, (3, self._ntbl))
        self._list = [self._ntbl, [3, self._ntbl]]
        self._dictionary = {
            "key1": self._ntbl,
            "key2": {"nested_key1": 3, "nested_key2": self._ntbl},
        }

    def test_tabulate_tuple(self):
        ntbl = tabulate(self._tuple)
        reflist = [
            (0, (3, 0)),
            ("1", (3, "1")),
            (2, (3, 2)),
            ("three", (3, "three")),
            (4, (3, 4)),
            ("5", (3, "5")),
        ]
        expected_ntbl = NTable(reflist, self._refmap)
        assert_ntable_equivalent(ntbl, expected_ntbl)

    def test_tabulate_list(self):
        ntbl = tabulate(self._list)
        reflist = [
            [0, [3, 0]],
            ["1", [3, "1"]],
            [2, [3, 2]],
            ["three", [3, "three"]],
            [4, [3, 4]],
            ["5", [3, "5"]],
        ]
        expected_ntbl = NTable(reflist, self._refmap)
        assert_ntable_equivalent(ntbl, expected_ntbl)

    def test_tabulate_dict(self):
        ntbl = tabulate(self._dictionary)
        reflist = [
            {"key1": 0, "key2": {"nested_key1": 3, "nested_key2": 0}},
            {"key1": "1", "key2": {"nested_key1": 3, "nested_key2": "1"}},
            {"key1": 2, "key2": {"nested_key1": 3, "nested_key2": 2}},
            {
                "key1": "three",
                "key2": {"nested_key1": 3, "nested_key2": "three"},
            },
            {"key1": 4, "key2": {"nested_key1": 3, "nested_key2": 4}},
            {"key1": "5", "key2": {"nested_key1": 3, "nested_key2": "5"}},
        ]
        expected_ntbl = NTable(reflist, self._refmap)
        assert_ntable_equivalent(ntbl, expected_ntbl)


if __name__ == "__main__":
    unittest.main()



import unittest


from tapr.main.conversion import ntable
from tapr.main.tabularization import tabularize
from tapr.main.utils import NULL
from tests.testing_utils import assert_ntable_equivalent


class TestTabularize(unittest.TestCase):
    def setUp(self):
        self._ntbl_a = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
            }
        )
        self._ntbl_b = ntable({"col1": "c1", "col2": "c2"}, dims=("dim1",))
        self._ntbl_c = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
                "row3": {"col1": "r3c1", "col2": "r3c2"},
            }
        )

    def test_tabularize_good(self):
        def func(a, b):
            return a + b

        dfunc = tabularize(func)
        result = dfunc(self._ntbl_a, self._ntbl_b)
        expected_result = ntable(
            {
                "row1": {"col1": "r1c1c1", "col2": "r1c2c2"},
                "row2": {"col1": "r2c1c1", "col2": "r2c2c2"},
            }
        )
        assert_ntable_equivalent(result, expected_result)

    def test_tabularize_bad(self):
        def func(a, b):
            return a + b

        dfunc = tabularize(func)
        result = dfunc(self._ntbl_a, self._ntbl_c)
        expected_result = ntable(
            {
                "row1": {"col1": "r1c1r1c1", "col2": "r1c2r1c2"},
                "row2": {"col1": "r2c1r2c1", "col2": "r2c2r2c2"},
                "row3": {"col1": NULL(), "col2": NULL()},
            }
        )
        assert_ntable_equivalent(result, expected_result)


if __name__ == "__main__":
    unittest.main()



import operator as op
import numpy as np
import pandas as pd
import datetime as dt

TAPR_RESERVED_KEYWORD = "__TAPR_RESERVED_COORDS__"

# TODO: Ensure that all operators are handled

UFUNC_TO_OP = {
    np.add: op.add,
    np.subtract: op.sub,
    np.multiply: op.mul,
    np.divide: op.truediv,
    np.equal: op.eq,
}

PRINTABLE_TYPES = {bool, int, float, pd.Timestamp, pd.Timedelta, dt.datetime}
PRINTABLE_TYPES.update(tuple(np.sctypeDict.values()))



import unittest

import numpy as np
import pandas as pd
import xarray as xr


from tapr.main.conversion import ntable
from tapr.main.ntable import NTable
from tests.testing_utils import assert_ntable_equivalent


def func(a, b):
    return a + b


class TestNTableStructure(unittest.TestCase):
    def setUp(self):
        self._ntbl_a = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
            }
        )
        self._ntbl_a_transpose = NTable(
            self._ntbl_a.reflist, self._ntbl_a.refmap.T
        )

    def test_flat(self):
        result = list(self._ntbl_a.struct.flat)
        expected = ["r1c1", "r1c2", "r2c1", "r2c2"]
        self.assertListEqual(result, expected)

    def test_T(self):
        result = self._ntbl_a.struct.T
        expected = self._ntbl_a_transpose
        assert_ntable_equivalent(result, expected)

    def test_getitem(self):
        # single index in boundaries
        result = self._ntbl_a.struct[0]
        expected = NTable(
            ["r1c1", "r1c2", "r2c1", "r2c2"], self._ntbl_a.refmap[0]
        )
        assert_ntable_equivalent(result, expected)

        # single index out of boundaries
        try:
            result = self._ntbl_a.struct[2]
        except IndexError:
            pass
        except Exception as e:
            self.fail(f"Unexpected exception {e} was raised")
        else:
            self.fail("Expected exception not raised")

        # dual index in boundaries
        result = self._ntbl_a.struct[0, 1]
        expected = NTable(
            ["r1c1", "r1c2", "r2c1", "r2c2"], self._ntbl_a.refmap[0, 1]
        )
        assert_ntable_equivalent(result, expected)

        # dual index out of boundaries
        try:
            result = self._ntbl_a.struct[2, 1]
        except IndexError:
            pass
        except Exception as e:
            self.fail(f"Unexpected exception {e} was raised")
        else:
            self.fail("Expected exception not raised")

    def test_setitem_scalar_value(self):
        # setitem with scalar value
        self._ntbl_a.struct[0] = "test"
        expected = NTable(
            ["test", "test", "r2c1", "r2c2"], self._ntbl_a.refmap
        )
        assert_ntable_equivalent(self._ntbl_a, expected)

    def test_setitem_ntable_value(self):
        # setitem with ntable value
        self._ntbl_a.struct[0] = self._ntbl_a.struct[1]
        expected = NTable(
            ["r2c1", "r2c2", "r2c1", "r2c2"], self._ntbl_a.refmap
        )
        assert_ntable_equivalent(self._ntbl_a, expected)


if __name__ == "__main__":
    unittest.main()



from collections.abc import Mapping
import operator as op
import functools as ft
import itertools as it

import numpy as np
import pandas as pd
import xarray as xr

from .utils import basic_refmap, NULL
from .processing import broadcast_tables, tabular_map
from .engines import StandardEngine
from .ttypes import STANDARD_TTYPE


def _tuplefy(*args):
    return args


def _tabulate_tuple(tup):
    tds = broadcast_tables(*tup)
    return tabular_map(_tuplefy, *tds)


def _listify(*args):
    return list(args)


def _tabulate_list(lst):
    tds = broadcast_tables(*lst)
    return tabular_map(_listify, *tds)


def _slicefy(*args):
    return slice(args[0], args[1], args[2])


def _tabulate_slice(slc):
    slice_tup = (slc.start, slc.stop, slc.step)
    tds = broadcast_tables(*slice_tup)
    return tabular_map(_slicefy, *tds)


def _dictify_tuple(tup):
    return {t[0]: t[1] for t in tup}


def _tabulate_dictionary(dictionary):
    items_tup = tuple(dictionary.items())
    compressed_tuple = tabulate(items_tup)
    result = tabular_map(_dictify_tuple, compressed_tuple)
    return result


def tabulate(con):
    if isinstance(con, dict):
        new_con = {}
        for k, v in con.items():
            try:
                new_con[k] = tabulate(v)
            except Exception:
                new_con[k] = v
        return _tabulate_dictionary(new_con)

    if isinstance(con, tuple):
        new_con = tuple()
        for item in con:
            try:
                new_con += (tabulate(item),)
            except Exception:
                new_con += (item,)
        return _tabulate_tuple(new_con)

    if isinstance(con, list):
        new_con = list()
        for item in con:
            try:
                new_con.append(tabulate(item))
            except Exception:
                new_con.append(item)
        return _tabulate_list(new_con)

    if isinstance(con, slice):
        slice_tup = (con.start, con.stop, con.step)
        new_slice_tup = tuple()
        for item in slice_tup:
            try:
                new_slice_tup += (tabulate(item),)
            except Exception:
                new_slice_tup += (item,)
        new_slice = slice(*new_slice_tup)
        return _tabulate_slice(new_slice)

    raise TypeError(f"unable to tabulate object of type {type(con)}")


def _mapping_depth(mapping):
    if isinstance(mapping, Mapping):
        return (
            max(map(_mapping_depth, mapping.values())) if mapping else 0
        ) + 1
    return 0


# def _flatten_mapping_values(mapping):
#     result = []
#     for item in mapping.values():
#         if not isinstance(item, Mapping):
#             result.append(item)
#         else:
#             result.extend(_flatten_mapping_values(item))
#     return result


def _get_nested_value(data, index, default=None):
    try:
        return ft.reduce(op.getitem, index, data)
    except (KeyError, IndexError):
        return default


def _extract_mapping_coords(mapping, dest=None, depth=0, dims=None):
    retflag = False
    if dest is None:
        # if dest is None we are in the top-most call
        maxdepth = _mapping_depth(mapping)
        dest = {i: [] for i in range(maxdepth)}
        retflag = True
    for k, v in mapping.items():
        if k not in dest[depth]:
            dest[depth].append(k)
        if isinstance(v, Mapping):
            _extract_mapping_coords(v, dest=dest, depth=depth + 1)
    if retflag:
        if dims is None:
            return {f"dim{k}": v for k, v in dest.items()}
        else:
            return {dims[k]: v for k, v in dest.items()}


def _mapping_to_ntable(mapping, dims=None, engine=None, ttype=None):
    from .ntable import NTable

    coords = _extract_mapping_coords(mapping, dims=dims)
    dmap = basic_refmap(coords, tuple(coords.keys()))
    data_keys = it.product(*coords.values())
    dlist = [_get_nested_value(mapping, index, NULL()) for index in data_keys]

    return NTable(dlist, dmap, engine=engine, ttype=ttype)


def _data_array_to_ntable(data_array, engine=None, ttype=None):
    from .ntable import NTable

    dlist = list(data_array.values.flat)
    dmap_values = np.arange(len(dlist)).reshape(*(data_array.shape))
    dmap = xr.DataArray(dmap_values, data_array.coords, data_array.dims)
    return NTable(dlist, dmap, engine=engine, ttype=ttype)


def _pandas_to_ntable(pds, dims=None, engine=None, ttype=None):
    from .ntable import NTable

    reflist = list(pds.values.flat)
    if isinstance(pds, pd.DataFrame):
        if dims is None:
            dims = ("rows", "cols")
        coords = {dims[0]: list(pds.index), dims[1]: list(pds.columns)}
    elif isinstance(pds, pd.Series):
        if dims is None:
            dims = ("rows",)
        coords = {dims[0]: list(pds)}
    refmap = basic_refmap(coords, dims)
    return NTable(reflist, refmap, engine=engine, ttype=ttype)


def ntable(obj, dims=None, engine=None, ttype=None):
    """
    Parameters
    ----------
    obj : Mapping, xr.DataArray, pd.DataFrame, NTable or container of NTable objects
        The object to create a NTable object from.
    dims : Sequence, optional
        The desired dimension names. If None, dim names will be dim0, dim1,
        dim2, etc. The default is None.
    engine : Engine, optional
        The engine to be used when doing computations. If None, a standard
        serial engine will be used. The default is None.
    ttype : set, optional
        Set of objects that the NTable may contain. If None, a set of common
        objects is used. The default is None.

    Raises
    ------
    TypeError
        DESCRIPTION.

    Returns
    -------
    ntbl : TYPE
        DESCRIPTION.

    """
    from .ntable import NTable

    if isinstance(obj, Mapping):
        ntbl = _mapping_to_ntable(obj, dims, engine=engine, ttype=ttype)
        return ntbl
    if isinstance(obj, xr.DataArray):
        ntbl = _data_array_to_ntable(obj, engine=engine, ttype=ttype)
        return ntbl
    if isinstance(obj, pd.DataFrame) or isinstance(obj, pd.Series):
        ntbl = _pandas_to_ntable(obj, dims, engine=engine, ttype=ttype)
        return ntbl
    if isinstance(obj, NTable):
        ntbl = NTable(obj.reflist, obj.refmap, engine=engine, ttype=ttype)
        return ntbl
    try:
        ntbl = tabulate(obj)
        return ntbl
    except TypeError:
        raise TypeError(f"Unable to convert {type(obj)} to NTable")



import plotly.express as module_

from tapr.main.tabularization import tabularize

# tabularize any callables in plotly.express
for name in dir(module_):
    value = getattr(module_,name)
    if callable(value):
        globals()[name] = tabularize(value)
    else:
        globals()[name] = value


import unittest

from tapr.main.utils import NULL
from tapr.main.qol import blank, sblank, cartograph
from tapr.main.conversion import ntable

from tests.testing_utils import assert_ntable_equivalent



class TestQol(unittest.TestCase):
    def test_blank(self):
        expected = ntable({"row0":{"col0":NULL(),"col1":NULL()}, "row1":{"col0":NULL(),"col1":NULL()}})
        result = blank(expected.struct.coords, expected.struct.dims)
        assert_ntable_equivalent(result, expected)

    def test_sblank(self):
        expected = ntable({"row0":{"col0":NULL(),"col1":NULL()}, "row1":{"col0":NULL(),"col1":NULL()}})
        # coords = {"rows":[f"row{i}" for i in range(2)], "cols":[f"col{i}" for i in range(2)]
        result = sblank(*expected.struct.shape)
        assert_ntable_equivalent(result, expected)

    def test_cartograph(self):
        expected = ntable({"row0":{"col0":("row0","col0"),"col1":("row0","col1")}, "row1":{"col0":("row1","col0"),"col1":("row1","col1")}})
        result = cartograph(expected)
        assert_ntable_equivalent(result, expected)





if __name__ == "__main__":
    unittest.main()



import pickle as pk
import json
import sys
import io

import numpy as np
import plotly as pl

from ..main.utils import NULL
from ..main.engines import StandardEngine, ProcessEngine, ThreadEngine

DEFAULT_TYPE_ID = "__TAPR_DEFAULT_TYPE__"

serializers = {}
deserializers = {DEFAULT_TYPE_ID:pk.loads}

def serializer(type_, id_):
    def decorator(func):
        serializers[type_] = (func, id_)

    return decorator

def deserializer(id_):
    def decorator(func):
        deserializers[id_] = func

    return decorator

def serialize(obj, allow_pickle=False):
    try:
        srlzr, type_id = serializers[type(obj)]
    except KeyError:
        srlzr, type_id = pk.dumps, DEFAULT_TYPE_ID

    if srlzr == pk.dumps:
        if not allow_pickle:
            raise ValueError(f"Non-pickle serializer for type {type(obj)} not found and allow_pickle flag was False. Either define a non-pickle serializer or set the flag to True.")
    return srlzr(obj), type_id

def deserialize(bytes_, id_, allow_pickle=False):
    try:
        dsrlzr = deserializers[id_]
    except KeyError:
        dsrlzr = pk.loads

    if dsrlzr == pk.loads:
        if not allow_pickle:
            raise ValueError(f"Non-pickle deserializer for type_id {id_} not found and allow_pickle flag was False. Either define a non-pickle deserializer or set the flag to True.")

    return dsrlzr(bytes_)


# Python

# python int
@serializer(int, "__py_int__")
def py_int_serializer(int_):
    return json.dumps(int_).encode()

@deserializer("__py_int__")
def py_int_deserializer(bytes_):
    return json.loads(bytes_.decode())

#python float
@serializer(float, "__py_float__")
def py_float_serializer(float_):
    return json.dumps(float_).encode()

@deserializer("__py_float__")
def py_float_deserializer(bytes_):
    return json.loads(bytes_.decode())

#python complex
@serializer(complex, "__py_complex__")
def py_complex_serializer(complex_):
    tup = (complex_.real, complex_.imag)
    return json.dumps(tup).encode()

@deserializer("__py_complex__")
def py_complex_deserializer(bytes_):
    tup = json.loads(bytes_.decode())
    return complex(*tup)

#python list
@serializer(list, "__py_list__")
def py_list_serializer(list_):
    return json.dumps(list_).encode()

@deserializer("__py_list__")
def py_list_deserializer(bytes_):
    return json.loads(bytes_.decode())

#python tuple
@serializer(tuple, "__py_tuple__")
def py_tuple_serializer(tuple_):
    return json.dumps(tuple_).encode()

@deserializer("__py_tuple__")
def py_tuple_deserializer(bytes_):
    return tuple(json.loads(bytes_.decode()))

#python set
@serializer(set, "__py_set__")
def py_set_serializer(set_):
    return json.dumps(tuple(set_)).encode()

@deserializer("__py_set__")
def py_set_deserializer(bytes_):
    return set(json.loads(bytes_.decode()))

#python dict
@serializer(dict, "__py_dict__")
def py_dict_serializer(dict_):
    return json.dumps(dict_).encode()

@deserializer("__py_dict__")
def py_dict_deserializer(bytes_):
    return json.loads(bytes_.decode())

#python string
@serializer(str, "__py_string__")
def py_string_serializer(string_):
    return string_.encode()

@deserializer("__py_string__")
def py_string_deserializer(bytes_):
    return bytes_.decode()

#python bytes
@serializer(bytes, "__py_bytes__")
def py_bytes_serializer(bytes_):
    return bytes_

@deserializer("__py_bytes__")
def py_bytes_deserializer(bytes_):
    return bytes_

#python bytearray
@serializer(bytearray, "__py_bytearray__")
def py_bytearray_serializer(bytearray_):
    return bytes(bytearray_)

@deserializer("__py_bytearray__")
def py_bytearray_deserializer(bytes_):
    return bytearray(bytes_)

# Numpy

# ndarray
@serializer(np.ndarray, "__numpy_ndarray__")
def numpy_ndarray_serializer(array):
    bytes_io = io.BytesIO(b"")
    np.save(bytes_io, array, allow_pickle=False)
    bytes_io.seek(0)
    return bytes_io.read()
@deserializer("__numpy_ndarray__")
def numpy_ndarray_deserializer(bytes_):
    bytes_io = io.BytesIO(bytes_)
    return np.load(bytes_io)


# Tapr

# NULL
@serializer(NULL, "__tapr_null__")
def tapr_null_serializer(null):
    return json.dumps(0).encode()

@deserializer("__tapr_null__")
def tapr_null_deserializer(bytes_):
    return NULL()


# StandardEngine
@serializer(StandardEngine, "__tapr_standard_engine__")
def tapr_standard_engine_serializer(standard_engine):
    return json.dumps(0).encode()

@deserializer("__tapr_standard_engine__")
def tapr_standard_engine_deserializer(bytes_):
    return StandardEngine()

# ThreadEngine
@serializer(ThreadEngine, "__tapr_thread_engine__")
def tapr_thread_engine_serializer(thread_engine):
    return json.dumps(thread_engine.threads).encode()

@deserializer("__tapr_thread_engine__")
def tapr_standard_engine_deserializer(bytes_):
    threads = json.loads(bytes_.decode())
    return ThreadEngine(threads)

# ProcessEngine
@serializer(ProcessEngine, "__tapr_process_engine__")
def tapr_process_engine_serializer(process_engine):
    return json.dumps(process_engine.processes).encode()

@deserializer("__tapr_process_engine__")
def tapr_standard_engine_deserializer(bytes_):
    processes = json.loads(bytes_.decode())
    return ProcessEngine(processes)


# Plotly

# Figure
@serializer(pl.graph_objs.Figure, "__plotly_figure__")
def plotly_figure_serializer(figure):
    return figure.to_json().encode()

@deserializer("__plotly_figure__")
def plotly_figure_deserializer(bytes_):
    json_ = bytes_.decode()
    dict_ = json.loads(json_)
    return pl.graph_objs.Figure(dict_)



from concurrent import futures as ft


class ProcessEngine:
    def __init__(self, processes):
        self._processes = processes

    @property
    def processes(self):
        return self._processes

    def __call__(self, func, *args):
        with ft.ProcessPoolExecutor(self._processes) as ex:
            return list(ex.map(func, *args))

    def __str__(self):
        return f"Process Engine\nProcesses: {self._processes}"

    def __repr__(self):
        return str(self)


class ThreadEngine:
    def __init__(self, threads):
        self._threads = threads

    @property
    def threads(self):
        return self._threads

    def __call__(self, func, *args):
        with ft.ThreadPoolExecutor(self._threads) as ex:
            return list(ex.map(func, *args))

    def __str__(self):
        return f"Thread Engine\nThreads: {self._threads}"

    def __repr__(self):
        return str(self)


class StandardEngine:
    def __init__(self):
        pass

    def __call__(self, func, *args):
        return list(map(func, *args))

    def __str__(self):
        return "Standard (serial) Engine"

    def __repr__(self):
        return str(self)
